(* Time-stamp: "convert.ml - alvaracu - Fri Oct 10 09:26:11 CEST 2003" *)

let (++) = Stream.iapp

exception Unknown_constant of string

let const s = 
  try Declare.global_absolute_reference (Libnames.path_of_string s) with _ -> 
  try Declare.global_reference (Names.id_of_string s) with _ ->
  raise (Unknown_constant s)

module Coq_eq = struct

  let t = lazy (const "eq")
  let mk a x y = Term.mkApp (Lazy.force t, [| a; x; y |])
  let lft e = (snd (Term.destApplication e)).(1)
  let rht e = (snd (Term.destApplication e)).(2)

end

let last_arg c = Util.array_last (snd (Term.destApplication c))

let make_elim x x_t gl =
  let case =
    if Termops.occur_term x (Tacmach.pf_concl gl) 
    then Indrec.make_case_dep
    else Indrec.make_case_gen in
  let ind, _ = Tacmach.pf_reduce_to_quantified_ind gl x_t in
  let sort = Tacticals.elimination_sort_of_goal gl in
  case (Tacmach.pf_env gl) (Tacmach.project gl) ind sort

let make_clause (x, x_t, x_bind) (elim, elim_bind) gl = 
  let wc, _ = Evar_refiner.startWalk gl in
  let x_t = 
    try snd (Tacmach.pf_reduce_to_quantified_ind gl x_t) 
    with Util.UserError _ -> x_t in
  let ind_cl = Clenv.make_clenv_binding wc (x, x_t) x_bind  in
  let elim_t = Evar_refiner.w_type_of wc elim in
  let elim_cl = Clenv.make_clenv_binding wc (elim, elim_t) elim_bind in 
  let ind_mv = Term.destMeta (last_arg elim_cl.Clenv.templval.Clenv.rebus) in
  Clenv.clenv_fchain ind_mv elim_cl ind_cl 

let make_residual elim_cl x_t gl = 
  let cl = Clenv.clenv_unique_resolver true elim_cl gl in
  let a = Clenv.clenv_instance_template cl in
  let _, d, e, _ = Term.destCase a in 
  let _, _, c = Term.destLambda d in 
  let beta x = Term.subst1 (Coq_eq.lft x) c in
  try beta x_t with _ -> beta (Tacmach.pf_type_of gl e)

let tac_in_concl (x, x_bind) x_t gl =
  let case = make_elim x x_t gl in  
  Pp.msgnl (Pp.str "conv .. " ++ Printer.prterm case);
  let case_cl = make_clause (x, x_t, x_bind) (case, x_bind) gl in
  Pp.msgnl (Pp.str "conv ?? " ++ Printer.prterm case_cl.Clenv.templval.Clenv.rebus);
  try 
    let c = make_residual case_cl x_t gl in
    Pp.msgnl (Pp.str "conv !! " ++ Printer.prterm c);
    Tactics.change_in_concl None c gl
  with _ -> Tacticals.tclFAIL 0 gl

let eq_in_concl e gl = 
  tac_in_concl (e, Rawterm.NoBindings) (Tacmach.pf_type_of gl e) gl

let in_concl x y gl =
  let h = Term.mkVar (Names.id_of_string "__convert_equality") in
  let e = Coq_eq.mk (Tacmach.pf_type_of gl x) x y in
  tac_in_concl (h, Rawterm.NoBindings) e gl

let eq_in_hyp h e gl = Tacticals.tclIDTAC gl
let in_hyp h x y gl = Tacticals.tclIDTAC gl


